#import "OITView.h"

char error[256];

@implementation OITView

#pragma mark Setup

- (id) initWithFrame:(NSRect)frame
{
	w = frame.size.width;
	h = frame.size.height;

	GLuint attribs[] = 
	{
		NSOpenGLPFANoRecovery,
		NSOpenGLPFAWindow,
		NSOpenGLPFAAccelerated,
		NSOpenGLPFADoubleBuffer,
		NSOpenGLPFAColorSize, 24,
		NSOpenGLPFAAlphaSize, 8,
		NSOpenGLPFADepthSize, 16,
		NSOpenGLPFAStencilSize, 8,
		NSOpenGLPFAAccumSize, 0,
		0
	};

	NSOpenGLPixelFormat* format = [[NSOpenGLPixelFormat alloc] initWithAttributes: (NSOpenGLPixelFormatAttribute*) attribs]; 
	
	if( !format ){ NSLog(@"No OpenGL pixel format"); exit(0); }
	
	return self = [super initWithFrame:frame pixelFormat:[format autorelease]];
}

- (void) awakeFromNib
{
	[[self window] makeFirstResponder:self];
	[[self window] setAcceptsMouseMovedEvents:YES];
} 

// Simple vertex and fragment shader load routine based off of the loader in GLSLShowpiece
- (GLhandleARB) loadVertexShader:(NSString*)vertexString fragmentShader:(NSString*)fragmentString
{
	const GLcharARB* vertex_string;
	const GLcharARB* fragment_string;
	
	GLhandleARB vertex_shader;
	GLhandleARB fragment_shader;
	GLhandleARB	program_object;
	
	GLint vertex_compiled = 0;
	GLint fragment_compiled = 0;
	
	GLint linked;
	
	if( vertexString )
	{
		vertex_shader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
		vertex_string = [vertexString cString];
		
		glShaderSourceARB(vertex_shader, 1, &vertex_string, NULL);
		glCompileShaderARB(vertex_shader);
		glGetObjectParameterivARB(vertex_shader, GL_OBJECT_COMPILE_STATUS_ARB, &vertex_compiled);
		
		if( !vertex_compiled ){ glGetInfoLogARB( vertex_shader, 256, 0, error ); NSLog( @"Log: %s\n", error ); }
	}
	else{ vertex_shader = NULL; vertex_compiled = 1; }
		
	if( fragmentString )
	{
		fragment_shader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
		fragment_string = [fragmentString cString];
		glShaderSourceARB(fragment_shader, 1, &fragment_string, NULL);
		glCompileShaderARB(fragment_shader);
		glGetObjectParameterivARB(fragment_shader, GL_OBJECT_COMPILE_STATUS_ARB, &fragment_compiled);
		
		if( !fragment_compiled ){ glGetInfoLogARB( fragment_shader, 256, 0, error ); NSLog( @"Log: %s\n", error ); }
	}
	else{ fragment_shader = NULL; fragment_compiled = 1; }
	
	if( !vertex_compiled || !fragment_compiled )
	{
		if( vertex_shader   ){ glDeleteObjectARB(vertex_shader);     vertex_shader = NULL; }
		if( fragment_shader ){ glDeleteObjectARB(fragment_shader); fragment_shader = NULL; }
		return NULL;
	}
	
	program_object = glCreateProgramObjectARB();
	
	if(   vertex_shader ){ glAttachObjectARB(program_object,   vertex_shader); glDeleteObjectARB(vertex_shader);   }
	if( fragment_shader ){ glAttachObjectARB(program_object, fragment_shader); glDeleteObjectARB(fragment_shader); }
	
	glLinkProgramARB(program_object);
	glGetObjectParameterivARB(program_object, GL_OBJECT_LINK_STATUS_ARB, &linked);
	
	if( !linked )
	{
		glGetInfoLogARB( program_object, 256, 0, error ); NSLog( @"Log: %s\n", error );
		
		glDeleteObjectARB(program_object);
		program_object = NULL;
	}
	
	return program_object;
}

// Set things up such as the layer textures which will hold the color and depth information for each layer
- (void) prepareOpenGL
{
	glClearColor( 0.0, 0.0, 0.0, 0.0 );
	
	glDepthFunc( GL_LEQUAL );
	glEnable( GL_DEPTH_TEST );
	
    glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
    glEnable( GL_BLEND );
	
	spherex = 0.0;
	spherey = 0.0;
	spherez = 0.0;
	
	glGenFramebuffersEXT(1, &fb);
	
	glGenTextures( NUMLAYERS, DT );
	glGenTextures( NUMLAYERS, layers );
	
	for( int i = 0; i < NUMLAYERS; i++ )
	{
		glBindTexture( GL_TEXTURE_RECTANGLE_EXT, DT[i] );
		glTexImage2D(GL_TEXTURE_RECTANGLE_EXT, 0, GL_DEPTH_COMPONENT, w, h, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_INT, 0L);
		  
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_COMPARE_MODE_ARB, GL_COMPARE_R_TO_TEXTURE_ARB); // Crucial!
		glTexParameterf(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_COMPARE_FUNC_ARB, GL_LEQUAL);
	
		glBindTexture(GL_TEXTURE_RECTANGLE_EXT, layers[i]);
		glTexImage2D(GL_TEXTURE_RECTANGLE_EXT, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, 0L);
		glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MIN_FILTER,  GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_MAG_FILTER,  GL_NEAREST);
		glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_S,  GL_CLAMP);
		glTexParameteri(GL_TEXTURE_RECTANGLE_EXT, GL_TEXTURE_WRAP_T,  GL_CLAMP);
	}
	
	peel = [self loadVertexShader:0L fragmentShader:[NSString stringWithContentsOfFile:[[NSBundle mainBundle]pathForResource:@"oit" ofType:@"frag"]]];
}

- (void) mouseMoved:(NSEvent*)event
{
	spherey -= [event deltaX];
	spherex -= [event deltaY];
	[self setNeedsDisplay:YES];
}

#pragma mark The Meat

// Do a full geometry pass
- (void) drawScene
{
	glViewport( 0, 0, w, h );

	glClear( GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( 90.0, 1.0, 1.0, 30.0 );

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glTranslatef( 0.0, 0.0, -15.0 );
	
	glRotatef( spherex, 1.0, 0.0, 0.0 );
	glRotatef( spherey, 0.0, 1.0, 0.0 );
	glRotatef( spherez, 0.0, 0.0, 1.0 );
	
	// Intersecting rectangles
	
	// Red rectangle
	/*glColor4f( 1.0, 0.0, 0.0, .5 );
	glBegin(GL_QUADS);
		glVertex3f( -5, -5,  5 );
		glVertex3f( -5,  5,  5 );
		glVertex3f(  5,  5, -5 );
		glVertex3f(  5, -5, -5 );
	glEnd();
	
	// Blue rectangle
	glColor4f( 0.0, 0.0, 1.0, .5 );
	glBegin(GL_QUADS);
		glVertex3f( -5, -5, -5 );
		glVertex3f( -5,  5, -5 );
		glVertex3f(  5,  5,  5 );
		glVertex3f(  5, -5,  5 );
	glEnd();*/
	
	// Overlapping triangle sides
	
	// Red rectangle
	glColor4f( 1.0, 0.0, 0.0, .5 );
	glBegin(GL_QUADS);
		glVertex3f(  2,  10,  5 );
		glVertex3f(  3,   9,  5 );
		glVertex3f( -8,  -3, -5 );
		glVertex3f( -9,  -2, -5 );
	glEnd();
	
	// Blue rectangle
	glColor4f( 0.0, 0.0, 1.0, .5 );
	glBegin(GL_QUADS);
		glVertex3f( -2,  9, -5 );
		glVertex3f( -1, 10, -5 );
		glVertex3f( 10, -2,  5 );
		glVertex3f(  9, -3,  5 );
	glEnd();
	
	// Green rectangle
	glColor4f( 0.0, 1.0, 0.0, .5 );
	glBegin(GL_QUADS);
		glVertex3f( -12, -1,  5 );
		glVertex3f( -12, .5,  5 );
		glVertex3f(  12, .5, -5 );
		glVertex3f(  12, -1, -5 );
	glEnd();
	
	[[self openGLContext] flushBuffer];
}

// Draw a frame showing the three color layers and the final correctly blended result
- (void) drawRect:(NSRect)rect
{
	// Disable blending so that pure unblended data gets written into every layer
	glDisable( GL_BLEND );

	// Do an initial geomatry pass to fill up the first layer with data.
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_EXT, layers[0], 0);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_RECTANGLE_EXT, DT[0], 0);
		[self drawScene];
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	
	// Now do additional passes for each other layer with the pixel shader attached and the previous layer's depth map bound as input through "uniform sampler2DRectShadow DT" in the shader.
	for( int i = 1; i < NUMLAYERS; i++ )
	{
		glUseProgramObjectARB(peel);
		
		glEnable(GL_TEXTURE_RECTANGLE_EXT);
		glBindTexture(GL_TEXTURE_RECTANGLE_EXT, DT[i-1]);
		
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb);
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_RECTANGLE_EXT, layers[i], 0);
			glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_RECTANGLE_EXT, DT[i], 0);
				[self drawScene];
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		glUseProgramObjectARB(NULL);
	}
	
	// Now that we have all the layers filled up it is time to draw the frame, first by showing each individual layer and then the final blended composite.
	glViewport( 0, 0, (GLsizei)rect.size.width, (GLsizei)rect.size.height );
	
	glClear( GL_COLOR_BUFFER_BIT + GL_DEPTH_BUFFER_BIT );
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho( -w/2, w/2, h/2, -h/2, -200.0, 200.0 );

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glEnable( GL_TEXTURE_RECTANGLE_EXT );
	glColor4f( 1.0, 1.0, 1.0, 1.0 );
	
	glTranslatef( -w/4, -h/4, 0.0 );

	// Draw the first layer in the upper left corner
	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, layers[0]);
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0, 0.0 ); glVertex3f( -w/4, -h/4, 0.0f );
		glTexCoord2f(   w, 0.0 ); glVertex3f(  w/4, -h/4, 0.0f );
		glTexCoord2f(   w,   h ); glVertex3f(  w/4,  h/4, 0.0f );
		glTexCoord2f( 0.0,   h ); glVertex3f( -w/4,  h/4, 0.0f );
	glEnd();
	glTranslatef( w/2, 0.0, 0.0 );
	
	// Draw the second layer in the upper right corner
	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, layers[1]);
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0, 0.0 ); glVertex3f( -w/4, -h/4, 0.0f );
		glTexCoord2f(   w, 0.0 ); glVertex3f(  w/4, -h/4, 0.0f );
		glTexCoord2f(   w,   h ); glVertex3f(  w/4,  h/4, 0.0f );
		glTexCoord2f( 0.0,   h ); glVertex3f( -w/4,  h/4, 0.0f );
	glEnd();
	glTranslatef( -w/2, h/2, 0.0 );
	
	// Draw the third layer in the lower left corner
	glBindTexture(GL_TEXTURE_RECTANGLE_EXT, layers[2]);
	glBegin( GL_QUADS );
		glTexCoord2f( 0.0, 0.0 ); glVertex3f( -w/4, -h/4, 0.0f );
		glTexCoord2f(   w, 0.0 ); glVertex3f(  w/4, -h/4, 0.0f );
		glTexCoord2f(   w,   h ); glVertex3f(  w/4,  h/4, 0.0f );
		glTexCoord2f( 0.0,   h ); glVertex3f( -w/4,  h/4, 0.0f );
	glEnd();
	glTranslatef( w/2, 0.0, 0.0 );
	
	// Now it is time to show what the final product looks like so we first turn blending back on.
	glEnable( GL_BLEND );
	
	// Draw each layer over each other in reverse order
	for( int j = NUMLAYERS - 1; j >= 0; j-- )
	{
		glBindTexture( GL_TEXTURE_RECTANGLE_EXT, layers[j] );
		
		glBegin( GL_QUADS );
			glTexCoord2f( 0.0, 0.0 ); glVertex3f( -w/4, -h/4, 0.0f );
			glTexCoord2f(   w, 0.0 ); glVertex3f(  w/4, -h/4, 0.0f );
			glTexCoord2f(   w,   h ); glVertex3f(  w/4,  h/4, 0.0f );
			glTexCoord2f( 0.0,   h ); glVertex3f( -w/4,  h/4, 0.0f );
		glEnd();
	}
	
	glDisable( GL_TEXTURE_RECTANGLE_EXT );
		
	[[self openGLContext] flushBuffer];
}

@end
